import logging
import os
from pathlib import Path

from mcipc.rcon.item import Item
from mcwb.itemlists import grab
from mcwb.types import Vec3
from mcwb.volume import Volume

from mciwb import Client
from mciwb.backup import Backup
from mciwb.server import HOST, MinecraftServer, backup_folder
from mciwb.threads import set_client
from tests.conftest import KEEP_SERVER, RCON_P, servers_folder

GITHUB_ACTIONS = "GITHUB_ACTIONS" in os.environ
backup_server_name = "mciwb_backup_server"
restore_server_name = "mciwb_restore_server"


def test_backup_restore(tmp_path: Path):
    """
    Test the backup and restore functionality.

    Backup a server, make a change to the world, restore the backup, and
    verify that the world has been restored.

    Requires use of client contexts so that we can restart clients.

    (TODO: need to look at why client.close does not work, as contexts were
    the only way to get this working)
    """
    BACKUP_RCON = 20600

    test_block = Vec3(100, 100, 100)

    # create a new world to backup from
    mc_backup = MinecraftServer(
        name=backup_server_name,
        rcon=BACKUP_RCON,
        password=RCON_P,
        server_folder=servers_folder / backup_server_name,
        world_type="flat",
        keep=KEEP_SERVER,
    )
    backup = Backup("backup_world", str(mc_backup.world), str(backup_folder))
    mc_backup.create(test=True)

    # make a change to the world which is to be backed up
    with Client(HOST, BACKUP_RCON, passwd=RCON_P) as client:
        set_client(client)
        result = client.setblock(test_block, Item.RED_CONCRETE.value)
        logging.info("setblock %s", result)
        # backup the world
        backup.backup()

    RESTORE_RCON = 20500
    # create a new world to restore into
    mc_restore = MinecraftServer(
        name=restore_server_name,
        rcon=RESTORE_RCON,
        password=RCON_P,
        server_folder=servers_folder / restore_server_name,
        world_type="flat",
        keep=KEEP_SERVER,
    )

    if GITHUB_ACTIONS:
        # GHA has issues with file locks on the stopped container
        # use mc docker's feature of creating from zip instead
        mc_restore.create(world_zip=backup.get_latest_zip(), test=True, force=True)
    else:
        mc_restore.create(test=True, force=True)
        mc_restore.stop()
        restore = Backup("restore_world", str(mc_restore.world), str(backup_folder))
        # restore the world and the start the server
        restore.restore()
        mc_restore.start()

    with Client(HOST, RESTORE_RCON, passwd=RCON_P) as client:
        set_client(client)
        # TODO mcwb should break out a getblock function from grab
        grab_volume = Volume.from_corners(test_block, test_block)
        restored_blocks = grab(client, grab_volume)

    # shut down the servers
    mc_restore.remove()
    mc_backup.remove()

    assert restored_blocks[0][0][0] == Item.RED_CONCRETE



import logging
from pathlib import Path
from typing import Optional

import typer
from IPython.terminal.embed import InteractiveShellEmbed

import mciwb
from mciwb import Iwb, __version__
from mciwb.server import HOST, MinecraftServer

cli = typer.Typer(add_completion=False)

server_name = "mciwb_server"
default_server_folder = Path.home() / server_name
def_pass = "default_pass"
def_port = 20100
def_world_type = "normal"


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@cli.callback()
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Print the version of ibek and exit",
    ),
    debug: bool = False,
):
    """Minecraft Interactive World Builder"""


def exception_handler(exception_type, exception, traceback):
    if logging.root.level > logging.DEBUG:
        logging.error("%s: %s", exception_type.__name__, exception)
    logging.debug("", exc_info=True)


def init_logging(debug: bool):
    if debug:
        logging.basicConfig(
            format="%(levelname)s: %(pathname)s:%(lineno)d %(funcName)s "
            "\n\t%(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(levelname)s:\t%(message)s",
            level=logging.INFO,
        )


@cli.command()
def shell(
    server: str = HOST,
    port: int = def_port,
    passwd: str = def_pass,
    player: str = "",
    debug: bool = False,
    test: bool = False,
):
    """
    Start an interactive iPython shell with a predefined world object
    connected to your Minecraft server.
    """
    init_logging(debug)

    try:
        mciwb.world = Iwb(server, port, passwd)

        if player:
            mciwb.world.add_player(player)
    except ConnectionRefusedError:
        logging.error("Could not connect to server")
        exit(1)
    finally:
        if mciwb.world is not None:
            mciwb.world.stop()

    # for quick access in the shell without qualifying the namespace
    world = mciwb.world

    logging.info("######### Starting Interactive Session ##########\n")

    # Prepare IPython shell with auto-reload so user code edits work immediately
    shell = InteractiveShellEmbed()
    shell.magic(r"%load_ext autoreload")
    shell.magic(r"%autoreload 2")
    # Also suppress traceback to avoid intimidating novice programmers
    ipython = shell.get_ipython()
    ipython._showtraceback = exception_handler  # type: ignore

    if test:
        # for testing just report the world object state
        print(mciwb.world)
    else:
        # enter iPython shell until users exits
        shell(colors="neutral")

    # Terminate all threads after interactive session exits
    world.stop()


@cli.command()
def start(
    password: str = def_pass,
    port: int = def_port,
    folder: Path = default_server_folder,
    world_type: str = def_world_type,
    debug: bool = False,
):
    """
    Start a minecraft server in the background. Default options automatically
    work with `mciwb shell`
    """
    init_logging(debug)

    if folder.exists():
        if not folder.is_dir():
            logging.error(f"{folder} is not a directory")
            exit(1)
        else:
            if port != def_port or world_type != def_world_type:
                logging.error(
                    f"server in {folder} already exists. "
                    "Cannot change settings on an existing server."
                )
                exit(1)
            else:
                logging.info(f"Launching existing Minecraft server in {folder}")
    else:
        logging.info(f"Creating new Minecraft server in {folder}")

    server = MinecraftServer(server_name, port, password, folder, world_type)
    server.create()


@cli.command()
def stop(
    debug: bool = False,
):
    """
    Start a minecraft server in the background. Default options automatically
    work with `mciwb shell`
    """
    init_logging(debug)

    MinecraftServer.stop_named(server_name)


if __name__ == "__main__":
    cli()



"""
Functions for launching and controlling a Minecraft server in a Docker container.
"""
import logging
import shutil
from datetime import datetime
from pathlib import Path
from time import sleep

from docker import from_env
from docker.models.containers import Container

from mciwb import Client

HOST = "localhost"

# the default locally mapped backup folder for minecraft data
backup_folder = Path.home() / "mciwb_backups"


class MinecraftServer:
    def __init__(
        self,
        name: str,
        rcon: int,
        password: str,
        server_folder: Path,
        world_type: str,
        keep: bool = True,
    ) -> None:
        self.rcon = rcon
        self.port = rcon + 1
        self.name = name
        self.password = password
        self.server_folder = server_folder
        self.world = self.server_folder / "world"
        self.world_type = world_type
        self.container = None
        self.keep = keep

    def wait_server(self):
        """
        Wait until the server is ready to accept rcon connections

        Multiple calls to this with container restarts require passing a count.
        This is how many times the wait code looks for the server to come online
        in the logs.

        This is required because the --since argument to the docker logs command
        fails to return any logs at all.
        """

        start_time: datetime = datetime.now()
        timeout = 100

        assert isinstance(self.container, Container)

        self.container.reload()
        if self.container.status != "running":
            logs = "\n".join(str(self.container.logs()).split(r"\n"))
            raise RuntimeError(f"minecraft server failed to start\n\n{logs}")

        logging.info("waiting for server to come online ...")
        for block in self.container.logs(stream=True):
            logging.debug(block.decode("utf-8").strip())
            if b"RCON running" in block:
                break
            elapsed = datetime.now() - start_time
            if elapsed.total_seconds() > timeout:
                raise RuntimeError("Timeout Starting minecraft")

        # wait until a connection is available
        for _ in range(10):
            try:
                with Client(HOST, self.rcon, passwd=self.password):
                    pass
                break
            except ConnectionRefusedError:
                sleep(2)
        else:
            raise RuntimeError("Timeout Starting minecraft")
        logging.info(f"Server {self.name} is online on port {self.port}")

    def stop(self):
        """
        Stop the minecraft server
        """
        assert isinstance(self.container, Container)
        logging.info(f"Stopping Minecraft Server {self.name} ...")
        self.container.stop()
        self.container.wait()

        logging.info(f"Stopped Minecraft Server {self.name} ...")

    def start(self):
        """
        Start the minecraft server
        """
        assert isinstance(self.container, Container)
        logging.info(f"Starting Minecraft Server {self.name} on port {self.port}...")
        self.container.start()
        self.wait_server()
        logging.info(f"Started Minecraft Server {self.name} ...")

    def remove(self, force=False):
        """
        Remove a minecraft server container
        """
        # set env var MCIWB_KEEP_SERVER to keep server alive for faster
        # repeated tests and viewing the world with a minecraft client
        if self.container and (not self.keep or force):
            logging.info(f"Removing Minecraft Server {self.name} ...")
            self.stop()
            self.container.remove()

    def create(self, world_zip=None, test=False, force=False) -> None:
        """
        Spin up a test minecraft server in a container

        world: a zip file to use as the world data
        """

        # create and launch minecraft container once per session
        docker_client = from_env()

        for container in docker_client.containers.list(all=True):
            assert isinstance(container, Container)
            if container.name == self.name:
                self.container = container
                if force:
                    self.remove(force=True)
                    break
                if container.status == "running":
                    logging.info(
                        f"Minecraft Server '{self.name}' "
                        f"already running on port {self.port}"
                    )
                    return
                else:
                    self.start()
                    return

        logging.info(
            f"Launching Minecraft Server '{self.name}' on port {self.port} ..."
        )

        env = {
            "EULA": "TRUE",
            "SERVER_PORT": self.port,
            "RCON_PORT": self.rcon,
            "ENABLE_RCON": "true",
            "RCON_PASSWORD": self.password,
            "SEED": 0,
            "LEVEL_TYPE": self.world_type,
            "OPS": "TransformerScorn",
            "MODE": "creative",
            "SPAWN_PROTECTION": "FALSE",
        }

        if test:
            env.update(
                {
                    "GENERATE_STRUCTURES": "false",
                    "SPAWN_ANIMALS": "false",
                    "SPAWN_MONSTERS": "false",
                    "SPAWN_NPCS": "false",
                    "VIEW_DISTANCE": " 5",
                    "LEVEL_TYPE": "FLAT",
                    "FORCE_WORLD_COPY": "TRUE",
                }
            )

            # offline mode disables OPS so dont use it if we are keeping the server
            # for local testing. But normally for running CI we want this option.
            if not self.keep:
                env["ONLINE_MODE"] = "FALSE"

        if world_zip:
            env["WORLD"] = str(world_zip)

        if not self.server_folder.exists():
            self.server_folder.mkdir(parents=True)
        elif test:
            shutil.rmtree(self.server_folder)
            self.server_folder.mkdir(parents=True)

        container = docker_client.containers.run(
            "itzg/minecraft-server",
            detach=True,
            environment=env,
            ports={f"{self.rcon}/tcp": self.rcon, f"{self.port}": self.port},
            restart_policy={"Name": "always" if self.keep else "no"},
            volumes={
                str(self.server_folder): {"bind": "/data", "mode": "rw"},
                str(backup_folder): {"bind": str(backup_folder), "mode": "rw"},
            },
            name=self.name,
        )

        self.container = container

        self.wait_server()

    @classmethod
    def stop_named(cls, name: str):
        """
        Stop a minecraft server by name
        """
        docker_client = from_env()
        for container in docker_client.containers.list(all=True):
            assert isinstance(container, Container)
            if container.name == name:
                logging.info(f"Stopping Minecraft Server {name} ...")
                container.stop()
                container.wait()
                container.remove()
                return
        logging.warning(f"Minecraft Server '{name}' not found")
